/***************************************************************************
                          codeparser.cpp  -  description
                             -------------------
    begin                : Die Jul 9 2002
    copyright            : (C) 2002 by Andr Simon
    email                : andre.simon1@gmx.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/*
 Basisklasse fuer HTMLCode, LatexCode, RtfCode und TexCode
 erkennt hervorzuhebende Strings in der Eingabe und setzt die
 in den ableitenden Klassen definierten Tags in die Ausgabe ein.
 Der Parser funktioniert mit Zustaenden, in die gewechselt wird,
 sobald ein Muster in der Eingabe entsprechend erkannt wurde.
*/

#include <iostream>
#include <string>
#include <sstream>
#include "codeparser.h"

using namespace std;

namespace highlight {

CodeParser::CodeParser()
{}

/* Konstruktor

Parameter: LanguageDefinition langDef:
           Definition der Programmiersprache, nach deren Daten geparst werden soll
           bool printLineNumbers:
           gibt an, ob Zeilennumerierung ausgegeben werden soll
           string styleName:
           Name des Farbthemes
           int numSpaces:
           Anzahl der Leerzeichen, die ein Tab ersetzen
*/
CodeParser::CodeParser( const LanguageDefinition &langDef,
                        bool printLineNumbers, const string &cssStyle, int numSpaces)
    : langInfo (langDef),
    showLineNumbers ( printLineNumbers),
    maskWs(false),
    fragmentOutput(false),
    styleName(cssStyle),
    numberSpaces(numSpaces),
    in(NULL),
    out(NULL),
    lineIndex(0),
    formatter(NULL)

{
   // aktuelles Token aus dem Eingabestrom
  token="";
  token.reserve(30);
  line.reserve(300);
  symbolString = langInfo.getSymbolString();
  docStyle = new DocumentStyle(styleName);
}


CodeParser::~CodeParser()
{
  delete formatter;
  delete docStyle;
}

void CodeParser::setFormatter( enum FormatStyle style)
{
  if (style!=DISABLED)
    formatter=new astyle::ASFormatter();
  
  switch (style) {
  case DISABLED:
      break;

  case ANSI:
      formatter->setBracketIndent(false);
      formatter->setSpaceIndentation(4);
      formatter->setBracketFormatMode(astyle::BREAK_MODE);
      formatter->setClassIndent(false);
      formatter->setSwitchIndent(false);
      formatter->setNamespaceIndent(false);
      break;

  case GNU:
      formatter->setBlockIndent(true);
      formatter->setSpaceIndentation(2);
      formatter->setBracketFormatMode(astyle::BREAK_MODE);
      formatter->setClassIndent(false);
      formatter->setSwitchIndent(false);
      formatter->setNamespaceIndent(false);
      break;

  case JAVA:
      formatter->setJavaStyle();
      formatter->setBracketIndent(false);
      formatter->setSpaceIndentation(4);
      formatter->setBracketFormatMode(astyle::ATTACH_MODE);
      formatter->setSwitchIndent(false);
      break;

  case KR:
      formatter->setBracketIndent(false);
      formatter->setSpaceIndentation(4);
      formatter->setBracketFormatMode(astyle::ATTACH_MODE);
      formatter->setClassIndent(false);
      formatter->setSwitchIndent(false);
      formatter->setNamespaceIndent(false);
      break;

  case LINUX:
      formatter->setBracketIndent(false);
      formatter->setSpaceIndentation(8);
      formatter->setBracketFormatMode(astyle::BDAC_MODE);
      formatter->setClassIndent(false);
      formatter->setSwitchIndent(false);
      formatter->setNamespaceIndent(false);
      break;
    }
}

void  CodeParser::setLanguageDefinition(const LanguageDefinition &langDef)
{
  langInfo=langDef;
  symbolString = langInfo.getSymbolString();
}

/** sucht vorwaerts ab Position searchPos Ziffer in s und liefert Integerwert
der gefundenen Zahl zurueck.
Im SymbolString stehen die den einzelnen Symbolen zugeordneten Konstanten
immer HINTER diesen Symbolen*/
enum State CodeParser::getState(const string &s, unsigned int searchPos)
{
  unsigned int i= searchPos+1, result=0;

  // nach Ziffer in s suchen
  do
    {
      i++;
    }
  while ((i<s.length()) && (! StringTools::isDigit(s[i] ))) ;

  // Zahl zusammensetzen
  while ((i<s.length()) && (StringTools::isDigit(s[i] )))
    {
      result = result *10 + (s[i]-'0');
      i++;
    }
  return ((result)? (enum State)result:_UNKNOWN);
}


/** liefert Wort ab Stringposition lineIndex-1 zurueck. */
string CodeParser::getIdentifier()
{
  lineIndex--;
  unsigned int startPos=lineIndex;
  bool negNumberEncountered=false;
  while (   (lineIndex < line.length())
            && ((( line[lineIndex] >= 'a') && (line[lineIndex] <= 'z')
                 ||   (line[lineIndex] >= 'A') && (line[lineIndex] <= 'Z'))
                ||   (line[lineIndex] == '-') 
                ||   (line[lineIndex] == '_')
                ||   (line[lineIndex] >= '0') && (line[lineIndex] <= '9') 
                       && !negNumberEncountered)
            ||   isAllowedChar(line[lineIndex]))
    {
      //falls auf ein '-' eine Zahl folgt, identifier-parsing abbrechen,
      //weil wahrscheinlich ein term x-zahl vorliegt
      negNumberEncountered=(line[lineIndex] == '-');
      lineIndex++;
    }
  return string(line, startPos, lineIndex - startPos);
}


/** liefert Zahl ab Stringposition lineIndex-1 zurueck.*/
string CodeParser::getNumber()
{
  lineIndex--;
  unsigned int startPos=lineIndex;
  char c= tolower(line[lineIndex]);
  bool isHex=false;
  
  while ( StringTools::isDigit(c)
          //mehr als 1 Punkt duerfte eh Compilierfehler geben ;)
          || (c == '.')
          //Minus als 1. Zeichen wird akzeptiert
          || (c == '-' && lineIndex == startPos)
          // Vorangestelltes 0x / 0X ist legal
          || (c=='x' && line[lineIndex-1]=='0')
          || ( c=='l' || c=='u'  )
          || ( isHex && (c>='a'  &&  c<='f')) ) 
    {
      if (c=='x' && line[lineIndex-1]=='0')
        isHex=true;
      lineIndex++;
      if (lineIndex < line.length()){
        c= tolower(line[lineIndex]);
      }
      else break;
    }
  return string(line,startPos, lineIndex- startPos);
}


/** fuegt am Anfang der neuen Zeile eine fortlaufende Nummer ein
Parameter: State envokingState
           Zustand des Aufrufers
*/
void CodeParser::insertLineNumber (State envokingState)
{
  if (showLineNumbers)
    {
      *out << styleTagClose[envokingState];
      *out << styleTagOpen[LINENUMBER] << StringTools::int2str(lineNumber, 4, spacer)
           << spacer << styleTagClose[LINENUMBER];
      *out << styleTagOpen[envokingState];
    }
}

/** gibt Zeilennummer der aktuellen Zeile zurck */
unsigned int CodeParser::getLineNumber()
{
  return lineNumber;
}


/** liefert den nchsten Character aus dem Inputstream */
unsigned char CodeParser::getInputChar()
{
  bool eol = false, eof=false;

  // Neue Zeile einlesen, falls lineIndex am Ende angelangt
  if (lineIndex == line.length())
    {
      terminatingChar=line[lineIndex-1];
      if (formatter==NULL)
        {  // Reformatierung nicht aktiviert
          eof = ! getline( *in, line);
        }
      // Eingabe vom Formatter lesen
      else
        {
          eof=!formatter->hasMoreLines();
          if (!eof)
            {
              line = formatter->nextLine();
            }
        }
      line=StringTools::trimRight(line);
      eol=true;
      lineIndex=0;
      lineNumber++;
    }

  if (eof)
    return '\0';
  if (eol)
    return '\n';
  return line[lineIndex++];
}

unsigned char CodeParser::getLastChar(){
  if (lineIndex<2 ) return '\0';
  return line[lineIndex-2];
}

/** gibt naechsten Character der Zeile zurueck, der kein Whitespace ist
lineIndex bleibt unveraendert*/
unsigned char CodeParser::getNextNonWs()
{
  unsigned char c;
  int index=lineIndex;
  do
    {
      c=line[index++];

    }
  while (StringTools::isWhitespace(c));
  return c;
}


/** liefert den Zustand zurueck, der durch das aktuelle Token eingeleitet wird.
ist token leer, wird in der aktuellen Eingabezeile gesucht,
sonst wird der interne Zeilenindex zurueckgesetzt, um das Token erneut zu parsen
Parameter: bool lastStateWasNumber:
           true, falls der letzte Zustand gleich Number war (zum Erkennen von Termen wie a-b)
Return:    naechsten Zustand
           token: Erkanntes Token, das den neuen Zustand einleitet
*/
enum State CodeParser::getCurrentState (bool lastStateWasNumber)
{
  //Kopie von Keywords/Typen bei nicht-casesensitiven Sprachen
  string caseSucksToken;

  unsigned char c;

  // Falls token leer, nchsten Character bestimmen
  if (token.length()==0)
    {
      c=getInputChar();
    }
  // sonst den Token nochmals untersuchen
  else
    {
      lineIndex-= (token.length()-1); //Stringindex zurcksetzen
      c=token.at(0);
    }

  if (c=='\n')
    return _EOL;   // End of line

  if (c=='\0')
    return _EOF;   // End of file

  // Whitespace erkannt; Tabs durch Spaces ersetzen, falls -t Option gesetzt wurde
  if (StringTools::isWhitespace(c))
    {
      if (numberSpaces && (c=='\t'))
        {
          for (int space=0; space < numberSpaces; space++)
            {
              token += spacer;
            }
        }
      else
        {
          if (c=='\t')
            token += c;
          else
            token += spacer;
        }
      return _WS;
    }

  // Zahl erkannt, Zahlen MUESSEN vor dem Suchen im Symbolstring geparst werden,
  // da sie sonst im Symbolstring gefunden werden!
  //falls c Ziffer oder '-' ist, Zahl filtern
  if ((StringTools::isDigit(c)) || langInfo.isNumberPrefix(c)
      //testen, ob '-' zu Zahl gehoehrt oder zu einem Term a-b:
      || (    (  c == '-')
              // falls vor derm '-' eine Zahl steht, gehoert '-' zu einem Term a-b
              && (!lastStateWasNumber)
              && StringTools::isDigit(getNextNonWs())) )
    {
      if (langInfo.isNumberPrefix(c))
        {
          token = c;
          lineIndex++;
          token += getNumber();
        }
      else
        {
          token = getNumber();
        }

      return NUMBER;
    }

  // nach Symbolen (z.B Kommentarbeginn, Direktiven) suchen.
  // Symbole werden VOR Keywords gesucht, da auch Buchstaben in Symbolen auftreten
  // koennen, zB in 4th
  //Leange des gefundenen Symbols in SymbolString:
  unsigned int symbTokenLength,
  symbolPos;
  bool found=false;

  //line[i] im symbolString suchen
  symbolPos = symbolString.find( c );

  while (  (symbolPos!= string::npos) && (!found))
    {
      //Laenge des Symbols bestimmen
      symbTokenLength = symbolString.find(' ', symbolPos) - symbolPos;

      //Symbol extrahieren
      token = symbolString.substr(symbolPos, symbTokenLength);

      // Abfrage nach Leerzeichen in SymbolString verhindert falsches
      // Erkennen von Symbolteilen:
      if (    token == line.substr(lineIndex-1, symbTokenLength)
              && (symbolString[symbolPos-1] == ' '))
        {
          found = true;
          lineIndex += (symbTokenLength-1);
        }
      //naechstes evtl. passende Symbol suchen
      else
        {
          symbolPos =
            symbolString.find_first_not_of(' ', symbolString.find(' ',symbolPos));
        }
    } //while

  if (found)
    {
      enum State foundState = getState(symbolString, symbolPos);
      if (langInfo.iAmForth()) {
        if (foundState==MULTI_LINE_COMMENT_BEGIN)
        {
          // 4th Workaround: Falls hinter MultilineBegin-Delimiter kein Whitespace
          // steht, ist es kein Kommentar
          c=getInputChar();
          token += c;
          // " (" korrekt erkennen
          if (!StringTools::isWhitespace(c))
            {
               foundState=((c=='\"')?STRING_END:STANDARD);
            }
        }
        // .(_ nicht als String erkennen
        if ( (token==".(") && (getInputChar()=='_')){
              foundState=STANDARD;
              lineIndex--;
         }
       }
       // Workaround: member attribute werden wie Tags behandelt,
       // solange kein eigener member attribute stil existiert
       if (isFirstNonWsChar()){
         if (foundState==MEMBER_ATTRIBUTE_BEGIN) return TAG_BEGIN;
       }
       if (foundState==MEMBER_ATTRIBUTE_END) return TAG_END;

      return foundState;
    }

  // Alphanumerisches Token parsen und als Keyword oder Type erkennen
  if (StringTools::isAlpha(c) || langInfo.isPrefix(c))
    {
      if (langInfo.isPrefix(c))
        {
          token = c;
          lineIndex++;
          token += getIdentifier();
        }
      else
        {
          token = getIdentifier();
        }

      //Falls Sprache nicht Case-sensitiv ist, Kopie von Token sichern
      //In keys/types-sets wird nach lowercase-Strings gesucht
      if (!langInfo.isCaseSensitive())
        {
          caseSucksToken = token;
          token = StringTools::lowerCase(token);
        }                      // Originalschreibweise wiederherstellen:

      bool isKeyword=langInfo.isKeyword(token),
           isType=false;

      if (!isKeyword) {
         isType= langInfo.isType(token);
      }

      // Originalschreibweise wiederherstellen:
      if ( ! langInfo.isCaseSensitive())
        {
         token = caseSucksToken;
         }

      if (isKeyword || isType)
        {
          return (isKeyword)?KEYWORD:TYPE;
        }

      return STANDARD;
    }

  // Character wurde nicht erkannt, in Standard-Zustand gehen
  token =c;
  return STANDARD;
}


/* maskiert Character aus s mit Escapesequenzen, die durch maskCharacter()
   definiert sind */
string CodeParser::maskString(const string & s)
{

  ostringstream ss;
  for (unsigned int i=0;i< s.length();i++)
    {
      ss << maskCharacter(s[i]);
    }
  return ss.str();
}

/* maskiert Character aus token mit Escapesequenzen */
void CodeParser::printMaskedToken()
{
  *out << maskString(token);
  token="";
}


/** gibt true zurck, falls c ein erlaubter Character innerhalb von Keyword
    oder Typbezeichner ist
*/
bool CodeParser::isAllowedChar(char c)
{
  string allowedChars=langInfo.getAllowedChars();
  return ( allowedChars.find(c)!= string::npos);
}

/* konvertiert inFileName in das angegebene Ausgabeformat und schreibt das Ergebnis
   nach outFileName
   Ist inFileName == "", wird cin verwendet;
   Ist outFileName == "", wird cout verwendet

*/

enum ParseError CodeParser::printOutput ( const string & inFileName, const string &outFileName,
                             bool fragment)
{
  fragmentOutput=fragment;
  if ( !assignInStream(inFileName))
     return BAD_INPUT;
  if (!assignOutStream(outFileName)){
      if (inFileName!="")
         delete in;
      return BAD_OUTPUT;
  }

  lineIndex = lineNumber = 0;

  if (formatter != NULL)
    {
      formatter->init(new astyle::ASStreamIterator(in));
    }
  if (! fragmentOutput)
    {
      printHeader(inFileName);
    }

  printBody();

  if (! fragmentOutput)
    {
      printFooter();
    }

  if (outFileName != "")
    {
      out->flush();
      ((ofstream*)out) -> close();
      delete out;
    }

   if (inFileName!= "")
     {
       ((ifstream*)in) -> close();
       delete in;
     }
   return PARSE_OK;
}


// Wenn mit der -i /-o - Option eine Ein-/Ausgabedatei angegeben
// wurde, ein filestream-Objekt erzeugen, sonst stdin/stdout zur
// Ein-/Ausgabe verwenden

bool CodeParser::assignInStream (const string& inFile)
{
  if (inFile != "")
    {
      in = new ifstream (inFile.c_str());
      if ( in->fail())
        {
          delete in;
          return false;
        }
    }
  else
    {
      in= &cin;
    }
  return true;
}

bool CodeParser::assignOutStream (const string& outFile)
{
  if (outFile != "")
    {
      out = new ofstream (outFile.c_str());
      if ( out->fail())
        {
          delete out;
          return false;
        }
    }
  else
    {
      out= &cout;
    }
  return true;
}

///////////////////////////////////////////////////////////////////////////////
////////////////////ZUSTAENDE PARSER///////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


/** Parst durch nicht als Sprachelemente erkannte Token,
die nicht hervorgehoben werden sollen.
Standard-State ist der Vaterzustand, der in die anderen
Zustnde verzweigt
*/
void CodeParser::processStandardState()
{
  // falls highlighting deaktiviert (zB bei txt.lang),
  // Eingabe nicht parsen, sondern lediglich Escapesequenzen einsetzen
  if (langInfo.highlightingDisabled()){
     string line;
     while (getline(*in, line)){
       *out << maskString(line) << newLineTag;
     }
     *out << flush;
  }
  else {
    // ID des erkannten Zustands (z.B Keyword, Type, Comment...)
    State state=STANDARD;


    // End of file-Flag
    bool eof=false;
    *out << styleTagOpen[STANDARD];
    do {
      // naechsten Zustand erkennen
      state= getCurrentState(state==NUMBER);

      switch(state)
        {                   // in entsprchenden Zustand verzweigen
        case TYPE:
          *out << styleTagClose[STANDARD];
          eof=processTrivialState(TYPE );
          *out << styleTagOpen[STANDARD];
          break;
        case KEYWORD:
          *out << styleTagClose[STANDARD];
          eof=processTrivialState(KEYWORD);
          *out << styleTagOpen[STANDARD];
          break;
        case NUMBER:
          *out << styleTagClose[STANDARD];
          eof=processTrivialState(NUMBER);
          *out << styleTagOpen[STANDARD];
          break;
        case MULTI_LINE_COMMENT_BEGIN:
          *out << styleTagClose[STANDARD];
          eof=processMultiLineCommentState( );
          *out << styleTagOpen[STANDARD];
          break;
        case SINGLE_LINE_COMMENT:
          *out << styleTagClose[STANDARD];
          eof=processSingleLineCommentState();
          *out << styleTagOpen[STANDARD];
          break;
        case STRING:
          *out << styleTagClose[STANDARD];
          eof=processStringState( STANDARD);
          *out << styleTagOpen[STANDARD];
          break;
        case DIRECTIVE_LINE:
          *out << styleTagClose[STANDARD];
          eof=processDirectiveState();
          *out << styleTagOpen[STANDARD];
          break;
        case TAG_BEGIN:
          *out << styleTagClose[STANDARD];
          eof=processTagState();
          *out << styleTagOpen[STANDARD];
          break;
        case ESC_CHAR:
          *out << styleTagClose[STANDARD];
          eof=processEscapeCharState();
          *out << styleTagOpen[STANDARD];
          break;
        case _EOL:               // ende der zeile erreicht
          *out <<newLineTag;
          insertLineNumber(STANDARD);
          break;
        case _EOF:
          eof=true;
          break;
        case _WS:
          processWsState();
          break;
        default:
          printMaskedToken();
          break;
        }
      }
    while (!eof);
    *out << styleTagClose[STANDARD];
    *out << flush;
  }
}


/** processTrivialState arbeitet einfache Zustaende ab, die ihrerseites keine
weiteren Zustaende aufrufen (Keywords, Types, Zahlen)

Parameter: int myState:
           der Zustand, in den gewechselt werden soll
return:    true falls Dateiende erreicht
*/
bool CodeParser::processTrivialState( enum State myState)
{
  enum State newState;
  bool eof=false,
           exitState=false; // wenn True, Zustand beenden

  *out << styleTagOpen[myState];
  do
    {
      printMaskedToken();
      newState= getCurrentState( myState==NUMBER);

      switch(newState)
        {
        case _WS: // verhindern, dass in default Fall eingestiegen wird, um mehrere
          // gleiche Token mit einem Tag zu formatieren
          processWsState();
          break;
        case _EOL: // Ende der Zeile erreicht
          *out << newLineTag;
          insertLineNumber(myState);
          break;
        case _EOF:  // Ende der Datei erreicht
          eof = true;
          break;
        default:        // Testen, ob ein neuer Zustand eingetreten ist
          exitState=(myState!=newState);
          break;
        } 
    }
  while ((!exitState) && (!eof));

  *out << styleTagClose[myState];
  return eof;
}


/** processMultiLineCommentState arbeitet mehrzeilige Kommentare ab

return:    true falls Dateiende erreicht
*/
bool CodeParser::processMultiLineCommentState()
{
  int commentCount=1;
  enum State newState;
  bool eof=false, exitState=false;

  *out << styleTagOpen[MULTI_LINE_COMMENT_BEGIN];
  do
    {
      printMaskedToken();
      newState= getCurrentState();

      switch(newState)
        {
          // verhindern, dass in default Fall eingestiegen wird,
          // um mehrere gleiche Token mit einem Tag zu formatieren
        case _WS:
          processWsState();
          break;
        case _EOL: // Ende der Zeile erreicht
          *out << newLineTag;
          insertLineNumber(MULTI_LINE_COMMENT_BEGIN);
          break;
        case _EOF:  // Ende der Datei erreicht
          eof = true;
          break;
        case MULTI_LINE_COMMENT_BEGIN:
          if (langInfo.allowNestedMLComments())
            {
              commentCount++;
            }
          break;
        case MULTI_LINE_COMMENT_END:
          commentCount--;
          if (! commentCount)
            {
              printMaskedToken();
              exitState=true;
            }
          break;
        default:
          break;
        }
    }
  while ((!exitState) && (!eof));

  *out << styleTagClose[MULTI_LINE_COMMENT_BEGIN];
  return eof;
}

/** processSingleLineCommentState arbeitet einzeilige Kommentare ab

return:    true falls Dateiende erreicht
*/
bool CodeParser::processSingleLineCommentState()
{
  int  newState;
  bool eof=false, exitState=false;

  *out << styleTagOpen[SINGLE_LINE_COMMENT];
  do
    {
      printMaskedToken();
      newState= getCurrentState();

      switch(newState)
        {
          // verhindern, dass in default Fall eingestiegen wird,
          // um mehrere gleiche Token mit einem Tag zu formatieren
        case _WS:
          processWsState();
          break;
        case _EOL: // Ende der Zeile erreicht
           printMaskedToken();
          exitState=true;
          *out << newLineTag;
          insertLineNumber(SINGLE_LINE_COMMENT);
          break;
        case _EOF:  // Ende der Datei erreicht
          eof = true;
          break;
        default:
          break;
        }
    }
  while ((!exitState) && (!eof));

  *out << styleTagClose[SINGLE_LINE_COMMENT];
  return eof;
}

/** processDirectiveState arbeitet einzeilige Direktiven ab

return:    true falls Dateiende erreicht
*/
bool CodeParser::processDirectiveState()
{
  enum State  newState;
  bool eof=false, exitState=false;

  *out << styleTagOpen[DIRECTIVE_LINE];
  do
    {
      printMaskedToken();
      newState= getCurrentState();
      switch(newState)
        {
        case _WS:
          processWsState();
          break;
        case DIRECTIVE_LINE_END:
           printMaskedToken();
          exitState=true;
          break;
        case _EOL: // Ende der Zeile erreicht
           printMaskedToken();
          // Forth beendet Directiven durch Delimiter
          if (!langInfo.iAmForth()
               && (terminatingChar!='\\')) // TODO: \ als LanguageDef.-Attribut
            {
              exitState=true;
            }
            *out << newLineTag;
            insertLineNumber(DIRECTIVE_LINE);
          break;
        case MULTI_LINE_COMMENT_BEGIN:
          *out<<styleTagClose[DIRECTIVE_LINE];
          eof= processMultiLineCommentState();
          *out << styleTagOpen[DIRECTIVE_LINE];
          break;
        case SINGLE_LINE_COMMENT:
          *out<<styleTagClose[DIRECTIVE_LINE];
          eof= processSingleLineCommentState();
          *out << styleTagOpen[DIRECTIVE_LINE];
          exitState=true;
          break;
        case STRING:
          *out << styleTagClose[DIRECTIVE_LINE];
          eof=processStringState(DIRECTIVE_LINE);
          *out << styleTagOpen[DIRECTIVE_LINE];
          break;
        case _EOF:  // Ende der Datei erreicht
          eof = true;
          break;
        default:
          break;
        }
    }
  while ((!exitState) && (!eof));

  *out << styleTagClose[DIRECTIVE_LINE];
  return eof;
}

/** processStringState arbeitet Strings ab

Parameter:  int  oldState:
           ist der Aufruferzustand
return:    true falls Dateiende erreicht
*/
bool CodeParser::processStringState(  enum State oldState)
{
  enum State  newState;
  bool eof=false, exitState=false,
       returnedFromOtherState=false ,
       disableEscCharHighlighting=false;
  // Merken, mit welchem Stringsymbol String geoeffnet wurde
  string openStringDelimiter=token;
  if (langInfo.getSimplifiedStringPrefix()!='\0') {
     disableEscCharHighlighting=getLastChar()==langInfo.getSimplifiedStringPrefix();
  }
  *out << ((oldState==DIRECTIVE_LINE)?
           styleTagOpen[DIRECTIVE_STRING] : styleTagOpen[STRING]);
  do
    {
      // Falls letzter Zustand EscapeChar war, muss Token nochmal geparst werden
      if (!returnedFromOtherState)
        {
           printMaskedToken();
        }
      returnedFromOtherState=false;
      newState= getCurrentState();

      switch(newState)
        {
        case _WS:
          processWsState();
          break;
        case _EOL: // Ende der Zeile erreicht
          *out << newLineTag;
          insertLineNumber(STRING);
          break;
        case MULTI_LINE_COMMENT_END:

          // Bei Forth ist der StringEnd und der MultiLineEnd-Delimiter derselbe
          if (langInfo.iAmForth())
            {
              newState=  STRING_END;
            }
          else
            {
              printMaskedToken();
              break;
            }
        case STRING:
        case STRING_END:

          // Workaround: Strings, die mit ." beginnen, knnen auch mit ." enden
          if (langInfo.iAmForth() && (newState==STRING) && (token!="\"")){
            exitState= ( openStringDelimiter==token)
                       || (openStringDelimiter=="\""); // falls String mit " beginnt, hier beenden
          } else{
             // String beendet, wenn Abschlussdelimiter gleich dem Beginndelimiter
           if (langInfo.iAmForth() && token=="(\""){
              exitState=true;
           } else {
              exitState= langInfo.isMatchingStringDelim(openStringDelimiter,token);
           }
        }
          printMaskedToken();
          break;
        case ESC_CHAR:
              // bei Simplified Strings (C#) nicht hervorheben
             if (!disableEscCharHighlighting){
                *out << ((oldState==DIRECTIVE_LINE)?
                       styleTagClose[DIRECTIVE_STRING] : styleTagClose[STRING]);
                eof=processEscapeCharState();
                *out << ((oldState==DIRECTIVE_LINE)?
                       styleTagOpen[DIRECTIVE_STRING] : styleTagOpen[STRING]);
                returnedFromOtherState=true;
             }
          break;
        case TYPE:
          if (langInfo.typesWithinStringsEnabled()){ 
             *out << styleTagClose[STRING];
             eof=processTrivialState(TYPE);
             *out << styleTagOpen[STRING];
             returnedFromOtherState=true;
           } else {
             printMaskedToken();
           }
          break;            
        case _EOF:
          eof = true;
          break;
        default:
          printMaskedToken();
          break;
        }
    }
  while ((!exitState) && (!eof));

  *out << ((oldState==DIRECTIVE_LINE)?
           styleTagClose[DIRECTIVE_STRING] : styleTagClose[STRING]);
  return eof;
}


/** processEscapeCharState arbeitet Escape Character ab

return:    true falls Dateiende erreicht
*/
bool CodeParser::processEscapeCharState()
{
  enum State newState;
  bool eof=false, exitState=false;

  *out << styleTagOpen[ESC_CHAR];
  skipEscapeSequence();
  do
    {
      token="";
      newState= getCurrentState();

      switch(newState)
        {
        case _EOL: // Ende der Zeile erreicht
          *out << newLineTag;
          insertLineNumber(ESC_CHAR);
          break;
        case ESC_CHAR:
          skipEscapeSequence();
          break;
        case _WS:
          processWsState();
          break;
        case _EOF:  // Ende der Datei erreicht
          eof = true;
          break;
        default:
          exitState=true;
          break;
        }
    }
  while ((!exitState) && (!eof));

  *out << styleTagClose[ESC_CHAR];
  return eof;
}

void  CodeParser::skipEscapeSequence(){
  printMaskedToken();
  if (lineIndex >= line.length())
    return;

     char c=line.at(lineIndex);
       int charsToSkip=1;
       // Escape Sequenz /ooo Oktal, /x00 hex, /u00xx Java unicode
       if (StringTools::isDigit(c)  && lineIndex < line.length()-2){
         // \0 abfangen
         while (StringTools::isDigit(line.at(lineIndex+charsToSkip))
                && charsToSkip<4) {
            charsToSkip++;
         }
       }
       else if ( c=='x' &&  lineIndex < line.length()-2){
            charsToSkip=3;
       } else if ( c=='u' && lineIndex < line.length()-4){
         charsToSkip=5;
       }
       for (int i=0;i<charsToSkip;i++){
         *out <<maskCharacter(line.at(lineIndex++));
       }
}
/** processTagState arbeitet Tags ab

return:    true falls Dateiende erreicht
*/
bool CodeParser::processTagState()
{
  enum State  newState;
  bool eof=false, exitState=false, returnedFromOtherState=false;
  // Merken, mit welchem Stringsymbol String geoeffnet wurde

  *out << styleTagOpen[KEYWORD];
  do
    {
      if (!returnedFromOtherState )
        {
          printMaskedToken();
        }
      returnedFromOtherState = false;
      newState= getCurrentState();

      switch(newState)
        {
        case _WS:
          processWsState();
          break;
        case _EOL: // Ende der Zeile erreicht
          *out << newLineTag;
          insertLineNumber(KEYWORD);
          break;
        case TAG_END:
          printMaskedToken();
          exitState=true;
          break;
        case STRING:
          *out << styleTagClose[KEYWORD];
          eof=processStringState( KEYWORD);
          *out << styleTagOpen[KEYWORD];
          returnedFromOtherState = true;
          break;
        case ESC_CHAR:
          *out << styleTagClose[KEYWORD];
          eof=processEscapeCharState();
          *out << styleTagOpen[KEYWORD];
          returnedFromOtherState = true;
          break;
        case NUMBER:
          *out << styleTagClose[KEYWORD];
          eof=processTrivialState(NUMBER );
          *out << styleTagOpen[KEYWORD];
          returnedFromOtherState = true;
          break;
        case _EOF:
          eof = true;
          break;
        default:
          printMaskedToken();
          break;
        }
    }
  while ((!exitState) && (!eof));

  *out << styleTagClose[KEYWORD];

  return eof;
}


/** processWsState arbeitet Whitespaces ab

return:    true falls Dateiende erreicht
*/
bool CodeParser::processWsState()
{
  if (!maskWs)
    {
      *out << token;
      token="";
      return false;
    }

  int wscnt=0;
  lineIndex--;
  while (StringTools::isWhitespace(line[lineIndex]))
    {
      if (line[lineIndex]==' ')
        {
          wscnt++;
        }
      else
        {
          wscnt+=numberSpaces;
        }
      lineIndex++;
    }
  if (wscnt==1)
    {
      *out << spacer;
    }
  // Sonderbehandlung fuer LateX-Einrueckung bei Zeilenbeginn:
  else
    {
      *out << maskWsBegin << wscnt*2 << maskWsEnd;
    }
  token="";
  return false;
}

bool CodeParser::isFirstNonWsChar() {
  unsigned int i=lineIndex-1;
  while (i--){
   if (!StringTools::isWhitespace(line[i])){
     return false;
   }
  }
  return true;
}

/**  setzt Flagvariablen auf Defaultwerte */
void CodeParser::reset()
{
 lineIndex = lineNumber = 0;
 line="";
}
}
